import json
import unittest
from requests.exceptions import ConnectionError
from unittest.mock import patch, MagicMock, Mock, ANY

import dbt.adapters
import google.cloud.bigquery

from dbt.adapters.bigquery import BigQueryCredentials
from dbt.adapters.bigquery import BigQueryRelation
from dbt.adapters.bigquery.connections import BigQueryConnectionManager
from dbt.adapters.bigquery.retry import RetryFactory


class TestBigQueryConnectionManager(unittest.TestCase):
    def setUp(self):
        self.credentials = Mock(BigQueryCredentials)
        self.credentials.method = "oauth"
        self.credentials.job_retries = 1
        self.credentials.job_retry_deadline_seconds = 1
        self.credentials.scopes = tuple()
        self.credentials.job_execution_timeout_seconds = 1

        self.mock_client = Mock(google.cloud.bigquery.Client)

        self.mock_connection = MagicMock()
        self.mock_connection.handle = self.mock_client
        self.mock_connection.credentials = self.credentials

        self.connections = BigQueryConnectionManager(
            profile=Mock(credentials=self.credentials, query_comment=None),
            mp_context=Mock(),
        )
        self.connections.get_thread_connection = lambda: self.mock_connection

    @patch(
        "dbt.adapters.bigquery.retry.create_bigquery_client",
        return_value=Mock(google.cloud.bigquery.Client),
    )
    def test_retry_connection_reset(self, mock_client_factory):
        new_mock_client = mock_client_factory.return_value

        @self.connections._retry.create_reopen_with_deadline(self.mock_connection)
        def generate_connection_reset_error():
            raise ConnectionResetError

        assert self.mock_connection.handle is self.mock_client

        with self.assertRaises(ConnectionResetError):
            # this will always raise the error, we just want to test that the connection was reopening in between
            generate_connection_reset_error()

        assert self.mock_connection.handle is new_mock_client
        assert new_mock_client is not self.mock_client

    def test_is_retryable(self):
        _is_retryable = google.cloud.bigquery.retry._job_should_retry
        exceptions = dbt.adapters.bigquery.impl.google.cloud.exceptions
        internal_server_error = exceptions.InternalServerError("code broke")
        bad_request_error = exceptions.BadRequest("code broke")
        connection_error = ConnectionError("code broke")
        client_error = exceptions.ClientError("bad code")
        rate_limit_error = exceptions.Forbidden(
            "code broke", errors=[{"reason": "rateLimitExceeded"}]
        )
        service_unavailable_error = exceptions.ServiceUnavailable("service is unavailable")

        self.assertTrue(_is_retryable(internal_server_error))
        self.assertFalse(
            _is_retryable(bad_request_error)
        )  # this was removed after initially being included
        self.assertTrue(_is_retryable(connection_error))
        self.assertFalse(_is_retryable(client_error))
        self.assertTrue(_is_retryable(rate_limit_error))
        self.assertTrue(_is_retryable(service_unavailable_error))

    def test_drop_dataset(self):
        mock_table = Mock()
        mock_table.reference = "table1"
        self.mock_client.list_tables.return_value = [mock_table]

        self.connections.drop_dataset("project", "dataset")

        self.mock_client.list_tables.assert_not_called()
        self.mock_client.delete_table.assert_not_called()
        self.mock_client.delete_dataset.assert_called_once()

    @patch("dbt.adapters.bigquery.connections.QueryJobConfig")
    def test_query_and_results(self, MockQueryJobConfig):
        self.mock_client.query.return_value = Mock(job_id="1")
        self.connections._query_and_results(
            self.mock_connection,
            "sql",
            {"dry_run": True},
            job_id=1,
        )

        MockQueryJobConfig.assert_called_once()
        self.mock_client.query.assert_called_once_with(
            query="sql",
            job_config=MockQueryJobConfig(),
            job_id=1,
            job_retry=None,
            timeout=self.credentials.job_creation_timeout_seconds,
        )

    def test_copy_bq_table_appends(self):
        self._copy_table(write_disposition=dbt.adapters.bigquery.impl.WRITE_APPEND)
        self.mock_client.copy_table.assert_called_once_with(
            [self._table_ref("project", "dataset", "table1")],
            self._table_ref("project", "dataset", "table2"),
            job_config=ANY,
            retry=ANY,
        )
        args, kwargs = self.mock_client.copy_table.call_args
        self.assertEqual(
            kwargs["job_config"].write_disposition, dbt.adapters.bigquery.impl.WRITE_APPEND
        )

    def test_copy_bq_table_truncates(self):
        self._copy_table(write_disposition=dbt.adapters.bigquery.impl.WRITE_TRUNCATE)
        args, kwargs = self.mock_client.copy_table.call_args
        self.mock_client.copy_table.assert_called_once_with(
            [self._table_ref("project", "dataset", "table1")],
            self._table_ref("project", "dataset", "table2"),
            job_config=ANY,
            retry=ANY,
        )
        args, kwargs = self.mock_client.copy_table.call_args
        self.assertEqual(
            kwargs["job_config"].write_disposition, dbt.adapters.bigquery.impl.WRITE_TRUNCATE
        )

    def test_job_labels_valid_json(self):
        expected = {"key": "value"}
        labels = self.connections._labels_from_query_comment(json.dumps(expected))
        self.assertEqual(labels, expected)

    def test_job_labels_invalid_json(self):
        labels = self.connections._labels_from_query_comment("not json")
        self.assertEqual(labels, {"query_comment": "not_json"})

    def test_list_dataset_correctly_calls_lists_datasets(self):
        mock_dataset = Mock(dataset_id="d1")
        mock_list_dataset = Mock(return_value=[mock_dataset])
        self.mock_client.list_datasets = mock_list_dataset
        result = self.connections.list_dataset("project")
        self.mock_client.list_datasets.assert_called_once_with(
            project="project", max_results=10000, retry=ANY
        )
        assert result == ["d1"]

    def _table_ref(self, proj, ds, table):
        return self.connections.table_ref(proj, ds, table)

    def _copy_table(self, write_disposition):
        source = BigQueryRelation.create(database="project", schema="dataset", identifier="table1")
        destination = BigQueryRelation.create(
            database="project", schema="dataset", identifier="table2"
        )
        self.connections.copy_bq_table(source, destination, write_disposition)



from dbt.tests.adapter.query_comment.test_query_comment import (
    BaseQueryComments,
    BaseMacroQueryComments,
    BaseMacroArgsQueryComments,
    BaseMacroInvalidQueryComments,
    BaseNullQueryComments,
    BaseEmptyQueryComments,
)


class TestQueryCommentsBigQuery(BaseQueryComments):
    pass


class TestMacroQueryCommentsBigQuery(BaseMacroQueryComments):
    pass


class TestMacroArgsQueryCommentsBigQuery(BaseMacroArgsQueryComments):
    pass


class TestMacroInvalidQueryCommentsBigQuery(BaseMacroInvalidQueryComments):
    pass


class TestNullQueryCommentsBigQuery(BaseNullQueryComments):
    pass


class TestEmptyQueryCommentsBigQuery(BaseEmptyQueryComments):
    pass



from dataclasses import dataclass
from typing import Any, Dict, Optional

from dbt.adapters.contracts.relation import (
    RelationConfig,
    ComponentName,
)
from google.cloud.bigquery import Table as BigQueryTable

from dbt.adapters.bigquery.relation_configs._base import BigQueryBaseRelationConfig
from dbt.adapters.bigquery.relation_configs._options import (
    BigQueryOptionsConfig,
    BigQueryOptionsConfigChange,
)
from dbt.adapters.bigquery.relation_configs._partition import (
    BigQueryPartitionConfigChange,
    PartitionConfig,
)
from dbt.adapters.bigquery.relation_configs._cluster import (
    BigQueryClusterConfig,
    BigQueryClusterConfigChange,
)


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class BigQueryMaterializedViewConfig(BigQueryBaseRelationConfig):
    """
    This config follow the specs found here:
    https://cloud.google.com/bigquery/docs/reference/standard-sql/data-definition-language#create_materialized_view_statement

    The following parameters are configurable by dbt:
    - table_id: name of the materialized view
    - dataset_id: dataset name of the materialized view
    - project_id: project name of the database
    - options: options that get set in `SET OPTIONS()` clause
    - partition: object containing partition information
    - cluster: object containing cluster information
    """

    table_id: str
    dataset_id: str
    project_id: str
    options: BigQueryOptionsConfig
    partition: Optional[PartitionConfig] = None
    cluster: Optional[BigQueryClusterConfig] = None

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "BigQueryMaterializedViewConfig":
        # required
        kwargs_dict: Dict[str, Any] = {
            "table_id": cls._render_part(
                ComponentName.Identifier, config_dict["table_id"]  # type:ignore
            ),
            "dataset_id": cls._render_part(
                ComponentName.Schema, config_dict["dataset_id"]  # type:ignore
            ),
            "project_id": cls._render_part(
                ComponentName.Database, config_dict["project_id"]  # type:ignore
            ),
            "options": BigQueryOptionsConfig.from_dict(config_dict["options"]),
        }

        # optional
        if partition := config_dict.get("partition"):
            kwargs_dict.update({"partition": PartitionConfig.parse(partition)})

        if cluster := config_dict.get("cluster"):
            kwargs_dict.update({"cluster": BigQueryClusterConfig.from_dict(cluster)})

        materialized_view: "BigQueryMaterializedViewConfig" = super().from_dict(
            kwargs_dict
        )  # type:ignore
        return materialized_view

    @classmethod
    def parse_relation_config(cls, relation_config: RelationConfig) -> Dict[str, Any]:
        config_dict = {
            "table_id": relation_config.identifier,
            "dataset_id": relation_config.schema,
            "project_id": relation_config.database,
            # despite this being a foreign object, there will always be options because of defaults
            "options": BigQueryOptionsConfig.parse_relation_config(relation_config),
        }

        # optional
        if relation_config.config and "partition_by" in relation_config.config:
            config_dict.update({"partition": PartitionConfig.parse_model_node(relation_config)})

        if relation_config.config and "cluster_by" in relation_config.config:
            config_dict.update(
                {"cluster": BigQueryClusterConfig.parse_relation_config(relation_config)}
            )

        return config_dict

    @classmethod
    def parse_bq_table(cls, table: BigQueryTable) -> Dict[str, Any]:
        config_dict = {
            "table_id": table.table_id,
            "dataset_id": table.dataset_id,
            "project_id": table.project,
            # despite this being a foreign object, there will always be options because of defaults
            "options": BigQueryOptionsConfig.parse_bq_table(table),
        }

        # optional
        if table.time_partitioning or table.range_partitioning:
            config_dict.update({"partition": PartitionConfig.parse_bq_table(table)})

        if table.clustering_fields:
            config_dict.update({"cluster": BigQueryClusterConfig.parse_bq_table(table)})

        return config_dict


@dataclass
class BigQueryMaterializedViewConfigChangeset:
    options: Optional[BigQueryOptionsConfigChange] = None
    partition: Optional[BigQueryPartitionConfigChange] = None
    cluster: Optional[BigQueryClusterConfigChange] = None

    @property
    def requires_full_refresh(self) -> bool:
        return any(
            {
                self.options.requires_full_refresh if self.options else False,
                self.partition.requires_full_refresh if self.partition else False,
                self.cluster.requires_full_refresh if self.cluster else False,
            }
        )

    @property
    def has_changes(self) -> bool:
        return any(
            {
                self.options if self.options else False,
                self.partition if self.partition else False,
                self.cluster if self.cluster else False,
            }
        )



import pytest
import os
from dbt.tests.util import run_dbt

_MODEL_SQL = """
select 1 as id
"""

_INVALID_LOCATION = os.getenv("DBT_TEST_BIGQUERY_BAD_LOCATION", "northamerica-northeast1")
_VALID_LOCATION = os.getenv("DBT_TEST_BIGQUERY_INITIAL_LOCATION", "US")


class BaseBigQueryLocation:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model.sql": _MODEL_SQL,
        }


class TestBigqueryValidLocation(BaseBigQueryLocation):
    def test_bigquery_valid_location(self, project):
        results = run_dbt()
        for result in results:
            assert "US" == result.adapter_response["location"]


class TestBigqueryInvalidLocation(BaseBigQueryLocation):
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        outputs = {"default": dbt_profile_target}
        outputs["default"]["location"] = _INVALID_LOCATION
        yield
        outputs = {"default": dbt_profile_target}
        outputs["default"]["location"] = _VALID_LOCATION

    def test_bigquery_location_invalid(self, project):
        results = run_dbt()
        for result in results:
            assert "northamerica-northeast1" == result.adapter_response["location"]



import datetime
import os
from unittest import mock
import pytest

from dbt.tests.adapter.sample_mode.test_sample_mode import (
    BaseSampleModeTest,
)
from dbt.tests.util import run_dbt

now = datetime.datetime.now()
twelve_hours_ago = now - datetime.timedelta(hours=12)
two_days_ago = now - datetime.timedelta(days=2)

_input_model_sql = f"""
{{{{ config(materialized='table', event_time='event_time') }}}}
select 1 as id, cast('{two_days_ago.strftime('%Y-%m-%d %H:%M:%S-0')}' as timestamp) as event_time
UNION ALL
select 2 as id, cast('{twelve_hours_ago.strftime('%Y-%m-%d %H:%M:%S-0')}' as timestamp) as event_time
UNION ALL
select 3 as id, cast('{now.strftime('%Y-%m-%d %H:%M:%S-0')}' as timestamp) as event_time
"""


class TestBigQuerySampleMode(BaseSampleModeTest):
    @pytest.fixture(scope="class")
    def input_model_sql(self) -> str:
        return _input_model_sql

    @mock.patch.dict(os.environ, {"DBT_EXPERIMENTAL_SAMPLE_MODE": "True"})
    def test_sample_mode(self, project) -> None:
        _ = run_dbt(["run"])
        self.assert_row_count(
            project=project,
            relation_name="model_that_samples_input_sql",
            expected_row_count=3,
        )

        _ = run_dbt(["run", "--sample=1 day"])
        self.assert_row_count(
            project=project,
            relation_name="model_that_samples_input_sql",
            expected_row_count=2,
        )



from dbt.adapters.bigquery.catalogs._biglake_metastore import BigLakeCatalogIntegration
from dbt.adapters.bigquery.catalogs._info_schema import BigQueryInfoSchemaCatalogIntegration
from dbt.adapters.bigquery.catalogs._relation import BigQueryCatalogRelation



import os

import pytest

from dbt.tests.util import run_dbt

_QUOTA_PROJECT = os.getenv("BIGQUERY_TEST_ALT_DATABASE")


class TestNoQuotaProject:
    def test_no_quota_project(self, project):
        results = run_dbt()
        for result in results:
            assert None == result.adapter_response["quota_project"]


class TestQuotaProjectOption:
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        outputs = {"default": dbt_profile_target}
        outputs["default"]["quota_project"] = _QUOTA_PROJECT
        yield

    def test_quota_project_option(self, project):
        results = run_dbt()
        for result in results:
            assert _QUOTA_PROJECT == result.adapter_response["quota_project"]



from dbt.tests.adapter.empty.test_empty import BaseTestEmpty, BaseTestEmptyInlineSourceRef


class TestBigQueryEmpty(BaseTestEmpty):
    pass


class TestBigQueryEmptyInlineSourceRef(BaseTestEmptyInlineSourceRef):
    pass



from multiprocessing import get_context
from unittest import mock

import agate
import decimal
import string
import random
import re
import pytest
import unittest
from unittest.mock import patch, MagicMock, create_autospec

import dbt_common.dataclass_schema
import dbt_common.exceptions.base

import dbt.adapters
from dbt.adapters.bigquery.relation_configs import PartitionConfig
from dbt.adapters.bigquery import BigQueryAdapter, BigQueryRelation
from google.cloud.bigquery.table import Table
from dbt.adapters.bigquery.connections import _sanitize_label, _VALIDATE_LABEL_LENGTH_LIMIT
from dbt_common.clients import agate_helper
import dbt_common.exceptions
from dbt.context.query_header import generate_query_header_context
from dbt.contracts.files import FileHash
from dbt.contracts.graph.manifest import ManifestStateCheck
from dbt.context.providers import RuntimeConfigObject, generate_runtime_macro_context

from google.cloud.bigquery import AccessEntry

from .utils import (
    config_from_parts_or_dicts,
    inject_adapter,
    TestAdapterConversions,
    load_internal_manifest_macros,
    mock_connection,
)


def _bq_conn():
    conn = MagicMock()
    conn.get.side_effect = lambda x: "bigquery" if x == "type" else None
    return conn


class BaseTestBigQueryAdapter(unittest.TestCase):
    def setUp(self):
        self.raw_profile = {
            "outputs": {
                "oauth": {
                    "type": "bigquery",
                    "method": "oauth",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                },
                "service_account": {
                    "type": "bigquery",
                    "method": "service-account",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "keyfile": "/tmp/dummy-service-account.json",
                    "threads": 1,
                },
                "external_oauth_wif": {
                    "type": "bigquery",
                    "method": "external-oauth-wif",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "token_endpoint": {
                        "type": "entra",
                        "request_url": "https://example.com/token",
                        "request_data": "mydata",
                    },
                    "audience": "https://example.com/audience",
                },
                "loc": {
                    "type": "bigquery",
                    "method": "oauth",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "location": "Luna Station",
                    "priority": "batch",
                    "maximum_bytes_billed": 0,
                },
                "api_endpoint": {
                    "type": "bigquery",
                    "method": "oauth",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "api_endpoint": "https://localhost:3001",
                },
                "impersonate": {
                    "type": "bigquery",
                    "method": "oauth",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "impersonate_service_account": "dummyaccount@dbt.iam.gserviceaccount.com",
                },
                "oauth-credentials-token": {
                    "type": "bigquery",
                    "method": "oauth-secrets",
                    "token": "abc",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "location": "Luna Station",
                    "priority": "batch",
                    "maximum_bytes_billed": 0,
                },
                "oauth-credentials": {
                    "type": "bigquery",
                    "method": "oauth-secrets",
                    "client_id": "abc",
                    "client_secret": "def",
                    "refresh_token": "ghi",
                    "token_uri": "jkl",
                    "project": "dbt-unit-000000",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "location": "Luna Station",
                    "priority": "batch",
                    "maximum_bytes_billed": 0,
                },
                "oauth-no-project": {
                    "type": "bigquery",
                    "method": "oauth",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "location": "Solar Station",
                },
                "dataproc-serverless-configured": {
                    "type": "bigquery",
                    "method": "oauth",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "gcs_bucket": "dummy-bucket",
                    "compute_region": "europe-west1",
                    "submission_method": "serverless",
                    "dataproc_batch": {
                        "environment_config": {
                            "execution_config": {
                                "service_account": "dbt@dummy-project.iam.gserviceaccount.com",
                                "subnetwork_uri": "dataproc",
                                "network_tags": ["foo", "bar"],
                            }
                        },
                        "labels": {"dbt": "rocks", "number": "1"},
                        "runtime_config": {
                            "properties": {
                                "spark.executor.instances": "4",
                                "spark.driver.memory": "1g",
                            }
                        },
                    },
                },
                "dataproc-serverless-default": {
                    "type": "bigquery",
                    "method": "oauth",
                    "schema": "dummy_schema",
                    "threads": 1,
                    "gcs_bucket": "dummy-bucket",
                    "compute_region": "europe-west1",
                    "submission_method": "serverless",
                },
            },
            "target": "oauth",
        }

        self.project_cfg = {
            "name": "X",
            "version": "0.1",
            "project-root": "/tmp/dbt/does-not-exist",
            "profile": "default",
            "config-version": 2,
        }
        self.qh_patch = None

        @mock.patch("dbt.parser.manifest.ManifestLoader.build_manifest_state_check")
        def _mock_state_check(self):
            all_projects = self.all_projects
            return ManifestStateCheck(
                vars_hash=FileHash.from_contents("vars"),
                project_hashes={name: FileHash.from_contents(name) for name in all_projects},
                profile_hash=FileHash.from_contents("profile"),
            )

        self.load_state_check = mock.patch(
            "dbt.parser.manifest.ManifestLoader.build_manifest_state_check"
        )
        self.mock_state_check = self.load_state_check.start()
        self.mock_state_check.side_effect = _mock_state_check

    def tearDown(self):
        if self.qh_patch:
            self.qh_patch.stop()
        super().tearDown()

    def get_adapter(self, target) -> BigQueryAdapter:
        project = self.project_cfg.copy()
        profile = self.raw_profile.copy()
        profile["target"] = target
        config = config_from_parts_or_dicts(
            project=project,
            profile=profile,
        )
        adapter = BigQueryAdapter(config, get_context("spawn"))
        adapter.set_macro_resolver(load_internal_manifest_macros(config))
        adapter.set_macro_context_generator(generate_runtime_macro_context)
        adapter.connections.set_query_header(
            generate_query_header_context(config, adapter.get_macro_resolver())
        )

        self.qh_patch = patch.object(adapter.connections.query_header, "add")
        self.mock_query_header_add = self.qh_patch.start()
        self.mock_query_header_add.side_effect = lambda q: "/* dbt */\n{}".format(q)

        inject_adapter(adapter)
        return adapter


class TestBigQueryAdapterAcquire(BaseTestBigQueryAdapter):
    @patch(
        "dbt.adapters.bigquery.credentials._create_bigquery_defaults",
        return_value=("credentials", "project_id"),
    )
    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_oauth_no_project_validations(
        self, mock_open_connection, mock_get_bigquery_defaults
    ):
        adapter = self.get_adapter("oauth-no-project")
        mock_get_bigquery_defaults.assert_called_once()
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_oauth_validations(self, mock_open_connection):
        adapter = self.get_adapter("oauth")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch(
        "dbt.adapters.bigquery.credentials._create_bigquery_defaults",
        return_value=("credentials", "project_id"),
    )
    @patch(
        "dbt.adapters.bigquery.connections.BigQueryConnectionManager.open", return_value=_bq_conn()
    )
    def test_acquire_connection_dataproc_serverless(
        self, mock_open_connection, mock_get_bigquery_defaults
    ):
        adapter = self.get_adapter("dataproc-serverless-configured")
        mock_get_bigquery_defaults.assert_called_once()
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.ValidationException as e:
            self.fail("got ValidationException: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_service_account_validations(self, mock_open_connection):
        adapter = self.get_adapter("service_account")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_oauth_token_validations(self, mock_open_connection):
        adapter = self.get_adapter("oauth-credentials-token")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_oauth_credentials_validations(self, mock_open_connection):
        adapter = self.get_adapter("oauth-credentials")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_impersonated_service_account_validations(
        self, mock_open_connection
    ):
        adapter = self.get_adapter("impersonate")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_external_oauth_wif_validations(self, mock_open_connection):
        adapter = self.get_adapter("external_oauth_wif")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        except BaseException:
            raise

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_priority(self, mock_open_connection):
        adapter = self.get_adapter("loc")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")
            self.assertEqual(connection.credentials.priority, "batch")

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    @patch("dbt.adapters.bigquery.BigQueryConnectionManager.open", return_value=_bq_conn())
    def test_acquire_connection_maximum_bytes_billed(self, mock_open_connection):
        adapter = self.get_adapter("loc")
        try:
            connection = adapter.acquire_connection("dummy")
            self.assertEqual(connection.type, "bigquery")
            self.assertEqual(connection.credentials.maximum_bytes_billed, 0)

        except dbt_common.exceptions.base.DbtValidationError as e:
            self.fail("got DbtValidationError: {}".format(str(e)))

        mock_open_connection.assert_not_called()
        connection.handle
        mock_open_connection.assert_called_once()

    def test_cancel_open_connections_empty(self):
        adapter = self.get_adapter("oauth")
        self.assertEqual(len(list(adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_master(self):
        adapter = self.get_adapter("oauth")
        key = adapter.connections.get_thread_identifier()
        adapter.connections.thread_connections[key] = mock_connection("master")
        self.assertEqual(len(list(adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_single(self):
        adapter = self.get_adapter("oauth")
        master = mock_connection("master")
        model = mock_connection("model")
        key = adapter.connections.get_thread_identifier()

        adapter.connections.thread_connections.update({key: master, 1: model})
        self.assertEqual(len(list(adapter.cancel_open_connections())), 1)

    @patch("dbt.adapters.bigquery.clients.ClientOptions")
    @patch("dbt.adapters.bigquery.credentials._create_bigquery_defaults")
    @patch("dbt.adapters.bigquery.clients.BigQueryClient")
    def test_location_user_agent(self, MockClient, mock_auth_default, MockClientOptions):
        creds = MagicMock()
        mock_auth_default.return_value = (creds, MagicMock())
        adapter = self.get_adapter("loc")

        connection = adapter.acquire_connection("dummy")
        mock_client_options = MockClientOptions.return_value

        MockClient.assert_not_called()
        connection.handle
        MockClient.assert_called_once_with(
            "dbt-unit-000000",
            creds,
            location="Luna Station",
            client_info=HasUserAgent(),
            client_options=mock_client_options,
        )

    @patch("dbt.adapters.bigquery.clients.ClientOptions")
    @patch("dbt.adapters.bigquery.credentials._create_bigquery_defaults")
    @patch("dbt.adapters.bigquery.clients.BigQueryClient")
    def test_api_endpoint_settable(self, MockClient, mock_auth_default, MockClientOptions):
        """Ensure api_endpoint is set on ClientOptions and passed to BigQueryClient."""

        creds = MagicMock()
        mock_auth_default.return_value = (creds, MagicMock())
        mock_client_options = MockClientOptions.return_value

        adapter = self.get_adapter("api_endpoint")
        connection = adapter.acquire_connection("dummy")
        MockClient.assert_not_called()
        connection.handle

        MockClientOptions.assert_called_once()
        kwargs = MockClientOptions.call_args.kwargs
        assert kwargs.get("api_endpoint") == "https://localhost:3001"

        MockClient.assert_called_once()
        assert MockClient.call_args.kwargs["client_options"] is mock_client_options


class HasUserAgent:
    PAT = re.compile(r"dbt-bigquery-\d+\.\d+\.\d+((a|b|rc)\d+)?")

    def __eq__(self, other):
        compare = getattr(other, "user_agent", "")
        return bool(self.PAT.match(compare))


class TestConnectionNamePassthrough(BaseTestBigQueryAdapter):
    def setUp(self):
        super().setUp()
        self._conn_patch = patch.object(BigQueryAdapter, "ConnectionManager")
        self.conn_manager_cls = self._conn_patch.start()

        self._relation_patch = patch.object(BigQueryAdapter, "Relation")
        self.relation_cls = self._relation_patch.start()

        self.mock_connection_manager = self.conn_manager_cls.return_value
        self.mock_connection_manager.get_if_exists().name = "mock_conn_name"
        self.conn_manager_cls.TYPE = "bigquery"
        self.relation_cls.get_default_quote_policy.side_effect = (
            BigQueryRelation.get_default_quote_policy
        )

        self.adapter = self.get_adapter("oauth")

    def tearDown(self):
        super().tearDown()
        self._conn_patch.stop()
        self._relation_patch.stop()

    def test_get_relation(self):
        self.adapter.get_relation("db", "schema", "my_model")
        self.mock_connection_manager.get_bq_table.assert_called_once_with(
            "db", "schema", "my_model"
        )

    @patch.object(BigQueryAdapter, "check_schema_exists")
    def test_drop_schema(self, mock_check_schema):
        mock_check_schema.return_value = True
        relation = BigQueryRelation.create(database="db", schema="schema")
        self.adapter.drop_schema(relation)
        self.mock_connection_manager.drop_dataset.assert_called_once_with("db", "schema")

    def test_get_columns_in_relation(self):
        self.mock_connection_manager.get_bq_table.side_effect = ValueError
        self.adapter.get_columns_in_relation(
            MagicMock(database="db", schema="schema", identifier="ident"),
        )
        self.mock_connection_manager.get_bq_table.assert_called_once_with(
            database="db", schema="schema", identifier="ident"
        )


class TestBigQueryRelation(unittest.TestCase):
    def setUp(self):
        pass

    def test_view_temp_relation(self):
        kwargs = {
            "type": None,
            "path": {"database": "test-project", "schema": "test_schema", "identifier": "my_view"},
            "quote_policy": {"identifier": False},
        }
        BigQueryRelation.validate(kwargs)

    def test_view_relation(self):
        kwargs = {
            "type": "view",
            "path": {"database": "test-project", "schema": "test_schema", "identifier": "my_view"},
            "quote_policy": {"identifier": True, "schema": True},
        }
        BigQueryRelation.validate(kwargs)

    def test_table_relation(self):
        kwargs = {
            "type": "table",
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "generic_table",
            },
            "quote_policy": {"identifier": True, "schema": True},
        }
        BigQueryRelation.validate(kwargs)

    def test_external_source_relation(self):
        kwargs = {
            "type": "external",
            "path": {"database": "test-project", "schema": "test_schema", "identifier": "sheet"},
            "quote_policy": {"identifier": True, "schema": True},
        }
        BigQueryRelation.validate(kwargs)

    def test_invalid_relation(self):
        kwargs = {
            "type": "invalid-type",
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "my_invalid_id",
            },
            "quote_policy": {"identifier": False, "schema": True},
        }
        with self.assertRaises(dbt_common.dataclass_schema.ValidationError):
            BigQueryRelation.validate(kwargs)


class TestBigQueryInformationSchema(unittest.TestCase):
    def setUp(self):
        pass

    def test_replace(self):
        kwargs = {
            "type": None,
            "path": {"database": "test-project", "schema": "test_schema", "identifier": "my_view"},
            # test for #2188
            "quote_policy": {"database": False},
            "include_policy": {
                "database": True,
                "schema": True,
                "identifier": True,
            },
        }
        BigQueryRelation.validate(kwargs)
        relation = BigQueryRelation.from_dict(kwargs)
        info_schema = relation.information_schema()

        tables_schema = info_schema.replace(information_schema_view="__TABLES__")
        assert tables_schema.information_schema_view == "__TABLES__"
        assert tables_schema.include_policy.schema is True
        assert tables_schema.include_policy.identifier is False
        assert tables_schema.include_policy.database is True
        assert tables_schema.quote_policy.schema is True
        assert tables_schema.quote_policy.identifier is False
        assert tables_schema.quote_policy.database is False

        schemata_schema = info_schema.replace(information_schema_view="SCHEMATA")
        assert schemata_schema.information_schema_view == "SCHEMATA"
        assert schemata_schema.include_policy.schema is False
        assert schemata_schema.include_policy.identifier is True
        assert schemata_schema.include_policy.database is True
        assert schemata_schema.quote_policy.schema is True
        assert schemata_schema.quote_policy.identifier is False
        assert schemata_schema.quote_policy.database is False

        other_schema = info_schema.replace(information_schema_view="SOMETHING_ELSE")
        assert other_schema.information_schema_view == "SOMETHING_ELSE"
        assert other_schema.include_policy.schema is True
        assert other_schema.include_policy.identifier is True
        assert other_schema.include_policy.database is True
        assert other_schema.quote_policy.schema is True
        assert other_schema.quote_policy.identifier is False
        assert other_schema.quote_policy.database is False


class TestBigQueryAdapter(BaseTestBigQueryAdapter):
    def test_copy_table_materialization_table(self):
        adapter = self.get_adapter("oauth")
        adapter.connections = MagicMock()
        adapter.copy_table("source", "destination", "table")
        adapter.connections.copy_bq_table.assert_called_once_with(
            "source", "destination", dbt.adapters.bigquery.impl.WRITE_TRUNCATE
        )

    def test_copy_table_materialization_incremental(self):
        adapter = self.get_adapter("oauth")
        adapter.connections = MagicMock()
        adapter.copy_table("source", "destination", "incremental")
        adapter.connections.copy_bq_table.assert_called_once_with(
            "source", "destination", dbt.adapters.bigquery.impl.WRITE_APPEND
        )

    def test_parse_partition_by(self):
        adapter = self.get_adapter("oauth")

        with self.assertRaises(dbt_common.exceptions.base.DbtValidationError):
            adapter.parse_partition_by("date(ts)")

        with self.assertRaises(dbt_common.exceptions.base.DbtValidationError):
            adapter.parse_partition_by("ts")

        self.assertEqual(
            adapter.parse_partition_by(
                {
                    "field": "ts",
                }
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "date",
                "granularity": "day",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {
                    "field": "ts",
                    "data_type": "date",
                }
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "date",
                "granularity": "day",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "date", "granularity": "MONTH"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "date",
                "granularity": "month",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "date", "granularity": "YEAR"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "date",
                "granularity": "year",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "timestamp", "granularity": "HOUR"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "timestamp",
                "granularity": "hour",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "timestamp", "granularity": "MONTH"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "timestamp",
                "granularity": "month",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "timestamp", "granularity": "YEAR"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "timestamp",
                "granularity": "year",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "datetime", "granularity": "HOUR"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "datetime",
                "granularity": "hour",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "datetime", "granularity": "MONTH"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "datetime",
                "granularity": "month",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "data_type": "datetime", "granularity": "YEAR"}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "datetime",
                "granularity": "year",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

        self.assertEqual(
            adapter.parse_partition_by(
                {"field": "ts", "time_ingestion_partitioning": True, "copy_partitions": True}
            ).to_dict(omit_none=True),
            {
                "field": "ts",
                "data_type": "date",
                "granularity": "day",
                "time_ingestion_partitioning": True,
                "copy_partitions": True,
            },
        )

        # Invalid, should raise an error
        with self.assertRaises(dbt_common.exceptions.base.DbtValidationError):
            adapter.parse_partition_by({})

        # passthrough
        self.assertEqual(
            adapter.parse_partition_by(
                {
                    "field": "id",
                    "data_type": "int64",
                    "range": {"start": 1, "end": 100, "interval": 20},
                }
            ).to_dict(omit_none=True),
            {
                "field": "id",
                "data_type": "int64",
                "granularity": "day",
                "range": {"start": 1, "end": 100, "interval": 20},
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            },
        )

    def test_hours_to_expiration(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {"hours_to_expiration": 4}
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {
            "expiration_timestamp": "TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL 4 hour)",
        }
        actual = adapter.get_table_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_hours_to_expiration_temporary(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {"hours_to_expiration": 4}
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {
            "expiration_timestamp": ("TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL 12 hour)"),
        }
        actual = adapter.get_table_options(mock_config, node={}, temporary=True)
        self.assertEqual(expected, actual)

    def test_table_kms_key_name(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {"kms_key_name": "some_key"}
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {"kms_key_name": "'some_key'"}
        actual = adapter.get_table_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_view_kms_key_name(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {"kms_key_name": "some_key"}
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {}
        actual = adapter.get_view_options(mock_config, node={})
        self.assertEqual(expected, actual)

    def test_get_common_options_labels_merge(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "labels": {"existing_label": "value1"},
            "labels_from_meta": True,
            "meta": {"meta_label": "value2"},
        }
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {"labels": [("meta_label", "value2"), ("existing_label", "value1")]}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_labels_no_meta(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "labels": {"existing_label": "value1"},
            "labels_from_meta": True,
            "meta": {},
        }
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {"labels": [("existing_label", "value1")]}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_labels_no_labels_from_meta(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "labels": {"existing_label": "value1"},
            "labels_from_meta": False,
            "meta": {"meta_label": "value2"},
        }
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {"labels": [("existing_label", "value1")]}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_no_labels(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "labels_from_meta": True,
            "meta": {"meta_label": "value2"},
        }
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {"labels": [("meta_label", "value2")]}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_empty(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {}
        mock_config.get.side_effect = lambda name: config.get(name)

        expected = {}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_resource_tags(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "resource_tags": {"test-project/env": "dev", "test-project/team": "data"},
        }
        mock_config.get.side_effect = lambda name, default=None: config.get(name, default)

        expected = {"tags": [("test-project/env", "dev"), ("test-project/team", "data")]}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        # Compare as sets since tag order depends on dictionary ordering
        self.assertEqual(set(actual["tags"]), set(expected["tags"]))

    def test_get_common_options_resource_tags_empty(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "resource_tags": {},
        }
        mock_config.get.side_effect = lambda name, default=None: config.get(name, default)

        expected = {}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)

    def test_get_common_options_resource_tags_and_labels(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "labels": {"label_key": "label_value"},
            "resource_tags": {"test-project/tag_key": "tag_value"},
        }
        mock_config.get.side_effect = lambda name, default=None: config.get(name, default)

        expected_labels = [("label_key", "label_value")]
        expected_tags = [("test-project/tag_key", "tag_value")]
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)

        self.assertEqual(set(actual["labels"]), set(expected_labels))
        self.assertEqual(set(actual["tags"]), set(expected_tags))

    def test_get_common_options_resource_tags_none(self):
        adapter = self.get_adapter("oauth")
        mock_config = create_autospec(RuntimeConfigObject)
        config = {
            "resource_tags": None,
        }
        mock_config.get.side_effect = lambda name, default=None: config.get(name, default)

        expected = {}
        actual = adapter.get_common_options(mock_config, node={}, temporary=False)
        self.assertEqual(expected, actual)


class TestBigQueryFilterCatalog(unittest.TestCase):
    def test__catalog_filter_table(self):
        used_schemas = [["a", "B"], ["a", "1234"]]
        column_names = ["table_name", "table_database", "table_schema", "something"]
        rows = [
            ["foo", "a", "b", "1234"],  # include
            ["foo", "a", "1234", "1234"],  # include, w/ table schema as str
            ["foo", "c", "B", "1234"],  # skip
            ["1234", "A", "B", "1234"],  # include, w/ table name as str
        ]
        table = agate.Table(rows, column_names, agate_helper.DEFAULT_TYPE_TESTER)

        result = BigQueryAdapter._catalog_filter_table(table, used_schemas)
        assert len(result) == 3
        for row in result.rows:
            assert isinstance(row["table_schema"], str)
            assert isinstance(row["table_database"], str)
            assert isinstance(row["table_name"], str)
            assert isinstance(row["something"], decimal.Decimal)


class TestBigQueryAdapterConversions(TestAdapterConversions):
    def test_convert_text_type(self):
        rows = [
            ["", "a1", "stringval1"],
            ["", "a2", "stringvalasdfasdfasdfa"],
            ["", "a3", "stringval3"],
        ]
        agate_table = self._make_table_of(rows, agate.Text)
        expected = ["string", "string", "string"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_text_type(agate_table, col_idx) == expect

    def test_convert_number_type(self):
        rows = [
            ["", "23.98", "-1"],
            ["", "12.78", "-2"],
            ["", "79.41", "-3"],
        ]
        agate_table = self._make_table_of(rows, agate.Number)
        expected = ["int64", "float64", "int64"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_number_type(agate_table, col_idx) == expect

    def test_convert_boolean_type(self):
        rows = [
            ["", "false", "true"],
            ["", "false", "false"],
            ["", "false", "true"],
        ]
        agate_table = self._make_table_of(rows, agate.Boolean)
        expected = ["bool", "bool", "bool"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_boolean_type(agate_table, col_idx) == expect

    def test_convert_datetime_type(self):
        rows = [
            ["", "20190101T01:01:01Z", "2019-01-01 01:01:01"],
            ["", "20190102T01:01:01Z", "2019-01-01 01:01:01"],
            ["", "20190103T01:01:01Z", "2019-01-01 01:01:01"],
        ]
        agate_table = self._make_table_of(
            rows, [agate.DateTime, agate_helper.ISODateTime, agate.DateTime]
        )
        expected = ["datetime", "datetime", "datetime"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_datetime_type(agate_table, col_idx) == expect

    def test_convert_date_type(self):
        rows = [
            ["", "2019-01-01", "2019-01-04"],
            ["", "2019-01-02", "2019-01-04"],
            ["", "2019-01-03", "2019-01-04"],
        ]
        agate_table = self._make_table_of(rows, agate.Date)
        expected = ["date", "date", "date"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_date_type(agate_table, col_idx) == expect

    def test_convert_time_type(self):
        # dbt's default type testers actually don't have a TimeDelta at all.
        agate.TimeDelta
        rows = [
            ["", "120s", "10s"],
            ["", "3m", "11s"],
            ["", "1h", "12s"],
        ]
        agate_table = self._make_table_of(rows, agate.TimeDelta)
        expected = ["time", "time", "time"]
        for col_idx, expect in enumerate(expected):
            assert BigQueryAdapter.convert_time_type(agate_table, col_idx) == expect

    # The casing in this case can't be enforced on the API side,
    # so we have to validate that we have a case-insensitive comparison
    def test_partitions_match(self):
        table = Table.from_api_repr(
            {
                "tableReference": {
                    "projectId": "test-project",
                    "datasetId": "test_dataset",
                    "tableId": "test_table",
                },
                "timePartitioning": {"type": "DAY", "field": "ts"},
            }
        )
        partition_config = PartitionConfig.parse(
            {
                "field": "TS",
                "data_type": "date",
                "granularity": "day",
                "time_ingestion_partitioning": False,
                "copy_partitions": False,
            }
        )
        assert BigQueryAdapter._partitions_match(table, partition_config) is True


class TestBigQueryGrantAccessTo(BaseTestBigQueryAdapter):
    entity = BigQueryRelation.from_dict(
        {
            "type": None,
            "path": {"database": "test-project", "schema": "test_schema", "identifier": "my_view"},
            "quote_policy": {"identifier": False},
        }
    )

    def setUp(self):
        super().setUp()
        self.mock_dataset: MagicMock = MagicMock(name="GrantMockDataset")
        self.mock_dataset.access_entries = [AccessEntry(None, "table", self.entity)]
        self.mock_client: MagicMock = MagicMock(name="MockBQClient")
        self.mock_client.get_dataset.return_value = self.mock_dataset
        self.mock_connection = MagicMock(name="MockConn")
        self.mock_connection.handle = self.mock_client
        self.mock_connection_mgr = MagicMock(
            name="GrantAccessMockMgr",
        )
        self.mock_connection_mgr.get_thread_connection.return_value = self.mock_connection
        _adapter = self.get_adapter("oauth")
        _adapter.connections = self.mock_connection_mgr
        self.adapter = _adapter

    def test_grant_access_to_calls_update_with_valid_access_entry(self):
        a_different_entity = BigQueryRelation.from_dict(
            {
                "type": None,
                "path": {
                    "database": "another-test-project",
                    "schema": "test_schema_2",
                    "identifier": "my_view",
                },
                "quote_policy": {"identifier": True},
            }
        )
        grant_target_dict = {"dataset": "someOtherDataset", "project": "someProject"}
        self.adapter.grant_access_to(
            entity=a_different_entity,
            entity_type="view",
            role=None,
            grant_target_dict=grant_target_dict,
        )
        self.mock_client.update_dataset.assert_called_once()


@pytest.mark.parametrize(
    ["input", "output"],
    [
        ("ABC", "abc"),
        ("a c", "a_c"),
        ("a ", "a"),
    ],
)
def test_sanitize_label(input, output):
    assert _sanitize_label(input) == output


@pytest.mark.parametrize(
    "label_length",
    [64, 65, 100],
)
def test_sanitize_label_length(label_length):
    random_string = "".join(
        random.choice(string.ascii_uppercase + string.digits) for i in range(label_length)
    )
    assert len(_sanitize_label(random_string)) <= _VALIDATE_LABEL_LENGTH_LIMIT



import pytest
from dbt.tests.util import run_dbt_and_capture

_MODEL_SQL = """
select 1 as id
"""


class BaseBigQueryHoursToExpiration:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model.sql": _MODEL_SQL,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {"test": {"materialized": "table", "model": {"hours_to_expiration": "4"}}}
        }


class TestBigQueryHoursToExpiration(BaseBigQueryHoursToExpiration):
    def test_bigquery_hours_to_expiration(self, project):
        _, stdout = run_dbt_and_capture(["--debug", "run"])
        assert "expiration_timestamp=TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL 4 hour)" in stdout



from typing import Optional, Tuple

import pytest

from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.contracts.relation import RelationType
from dbt.tests.adapter.materialized_view.files import MY_TABLE, MY_VIEW
from dbt.tests.util import (
    get_connection,
    get_model_file,
    run_dbt,
    set_model_file,
)

from tests.functional.adapter.materialized_view_tests import _files


class BigQueryMaterializedViewMixin:
    @pytest.fixture(scope="class")
    def my_base_table(self, project) -> BaseRelation:
        """
        The base table for a materialized view needs to be partitioned in
        the same way as the materialized view. So if we want to create a partitioned
        materialized view, we need to partition the base table. This table is a
        select * on the seed table, plus a partition.
        """
        return project.adapter.Relation.create(
            identifier="my_base_table",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.Table,
        )

    @pytest.fixture(scope="class")
    def my_other_base_table(self, project) -> BaseRelation:
        """
        Following the sentiment of `my_base_table` above, if we want to alter the partition
        on the materialized view, we either need to update the partition on the base table,
        or we need a second table with a different partition.
        """
        return project.adapter.Relation.create(
            identifier="my_other_base_table",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.Table,
        )

    @pytest.fixture(scope="function", autouse=True)
    def setup(self, project, my_base_table, my_other_base_table, my_materialized_view):  # type: ignore
        run_dbt(["seed"])
        run_dbt(["run", "--full-refresh"])

        # the tests touch these files, store their contents in memory
        initial_model = get_model_file(project, my_materialized_view)

        yield

        # and then reset them after the test runs
        set_model_file(project, my_materialized_view, initial_model)
        project.run_sql(f"drop schema if exists {project.test_schema} cascade")

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": _files.MY_SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            "my_table.sql": MY_TABLE,
            "my_view.sql": MY_VIEW,
            "my_base_table.sql": _files.MY_BASE_TABLE,
            "my_other_base_table.sql": _files.MY_OTHER_BASE_TABLE,
            "my_materialized_view.sql": _files.MY_MATERIALIZED_VIEW,
        }

    @staticmethod
    def insert_record(project, table: BaseRelation, record: Tuple[int, int]) -> None:
        my_id, value = record
        project.run_sql(f"insert into {table} (id, value) values ({my_id}, {value})")

    @staticmethod
    def refresh_materialized_view(project, materialized_view: BaseRelation) -> None:
        sql = f"""
        call bq.refresh_materialized_view(
            '{materialized_view.database}.{materialized_view.schema}.{materialized_view.identifier}'
        )
        """
        project.run_sql(sql)

    @staticmethod
    def query_row_count(project, relation: BaseRelation) -> int:
        sql = f"select count(*) from {relation}"
        return project.run_sql(sql, fetch="one")[0]

    # look into syntax
    @staticmethod
    def query_relation_type(project, relation: BaseRelation) -> Optional[str]:
        with get_connection(project.adapter) as conn:
            table = conn.handle.get_table(
                project.adapter.connections.get_bq_table(
                    relation.database, relation.schema, relation.identifier
                )
            )
        return table.table_type.lower()



my_model_struct_wrong_data_type_sql = """
{{ config(materialized = "table") }}

select
  STRUCT(1 AS struct_column_being_tested, "test" AS another_struct_column) as a
"""

my_model_struct_correct_data_type_sql = """
{{ config(materialized = "table")}}

select
  STRUCT("test" AS struct_column_being_tested, "test" AS b) as a
"""

model_struct_data_type_schema_yml = """
version: 2
models:
  - name: contract_struct_wrong
    config:
      contract:
        enforced: true
    columns:
      - name: a.struct_column_being_tested
        data_type: string
      - name: a.b
        data_type: string

  - name: contract_struct_correct
    config:
      contract:
        enforced: true
    columns:
      - name: a.struct_column_being_tested
        data_type: string
      - name: a.b
        data_type: string
"""

my_model_double_struct_wrong_data_type_sql = """
{{ config(materialized = "table") }}

select
  STRUCT(
    STRUCT(1 AS struct_column_being_tested, "test" AS c) as b,
    "test" as d
    ) as a
"""

my_model_double_struct_correct_data_type_sql = """
{{ config(materialized = "table") }}

select
  STRUCT(
    STRUCT("test" AS struct_column_being_tested, "test" AS c) as b,
    "test" as d
    ) as a
"""

model_double_struct_data_type_schema_yml = """
version: 2
models:
  - name: contract_struct_wrong
    config:
      contract:
        enforced: true
    columns:
      - name: a.b.struct_column_being_tested
        data_type: string
      - name: a.b.c
        data_type: string
      - name: a.d
        data_type: string

  - name: contract_struct_correct
    config:
      contract:
        enforced: true
    columns:
      - name: a.b.struct_column_being_tested
        data_type: string
      - name: a.b.c
        data_type: string
      - name: a.d
        data_type: string
"""


my_model_struct_sql = """
{{
  config(
    materialized = "table"
  )
}}

select STRUCT("test" as nested_column, "test" as nested_column2) as id
"""


model_struct_schema_yml = """
version: 2
models:
  - name: my_model
    config:
      contract:
        enforced: true
    columns:
      - name: id.nested_column
        quote: true
        data_type: string
        description: hello
        constraints:
          - type: not_null
          - type: unique
      - name: id.nested_column2
        data_type: string
        constraints:
          - type: unique
"""



from dataclasses import dataclass
from typing import Any, Dict, FrozenSet, Optional

from dbt.adapters.relation_configs import RelationConfigChange
from dbt.adapters.contracts.relation import RelationConfig
from google.cloud.bigquery import Table as BigQueryTable
from typing_extensions import Self

from dbt.adapters.bigquery.relation_configs._base import BigQueryBaseRelationConfig


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class BigQueryClusterConfig(BigQueryBaseRelationConfig):
    """
    This config manages table options supporting clustering. See the following for more information:
        - https://docs.getdbt.com/reference/resource-configs/bigquery-configs#using-table-partitioning-and-clustering
        - https://cloud.google.com/bigquery/docs/reference/standard-sql/data-definition-language#clustering_column_list

    - fields: set of columns to cluster on
        - Note: can contain up to four columns
    """

    fields: FrozenSet[str]

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> Self:
        kwargs_dict = {"fields": config_dict.get("fields")}
        return super().from_dict(kwargs_dict)  # type:ignore

    @classmethod
    def parse_relation_config(cls, relation_config: RelationConfig) -> Dict[str, Any]:
        config_dict = {}

        if cluster_by := relation_config.config.extra.get("cluster_by"):  # type:ignore
            # users may input a single field as a string
            if isinstance(cluster_by, str):
                cluster_by = [cluster_by]
            config_dict.update({"fields": frozenset(cluster_by)})

        return config_dict

    @classmethod
    def parse_bq_table(cls, table: BigQueryTable) -> Dict[str, Any]:
        config_dict = {"fields": frozenset(table.clustering_fields)}
        return config_dict


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class BigQueryClusterConfigChange(RelationConfigChange):
    context: Optional[BigQueryClusterConfig]

    @property
    def requires_full_refresh(self) -> bool:
        return True



import pytest
from dbt.tests.util import run_dbt


_MODEL_SQL = """
select {{ dbt.string_literal('my multiline
string') }} as test
"""


class TestStringLiteralQuoting:
    @pytest.fixture(scope="class")
    def models(self):
        return {"my_model.sql": _MODEL_SQL}

    def test_string_literal_quoting(self, project):
        run_dbt()



import unittest

from dbt.adapters.bigquery.relation_configs._options import BigQueryOptionsConfig


class TestBigQueryOptionsConfig(unittest.TestCase):
    """Unit tests for BigQuery options config with tags"""

    def test_from_dict_with_tags(self):
        """Test creating BigQueryOptionsConfig from dict with tags"""
        config_dict = {
            "enable_refresh": True,
            "refresh_interval_minutes": 45,
            "description": "Test materialized view",
            "labels": {"env": "dev", "version": "1.0"},
            "tags": {"test-project/team": "data", "test-project/cost_center": "analytics"},
        }

        options = BigQueryOptionsConfig.from_dict(config_dict)

        # Verify all fields are correctly set
        self.assertTrue(options.enable_refresh)
        self.assertEqual(options.refresh_interval_minutes, 45)
        self.assertEqual(options.description, "Test materialized view")
        self.assertEqual(options.labels, {"env": "dev", "version": "1.0"})
        self.assertEqual(
            options.tags, {"test-project/team": "data", "test-project/cost_center": "analytics"}
        )

    def test_as_ddl_dict_with_tags(self):
        """Test generating DDL dict with tags"""
        options = BigQueryOptionsConfig(
            enable_refresh=True,
            refresh_interval_minutes=30,
            description="Test MV",
            labels={"env": "test"},
            tags={"test-project/team": "data"},
        )

        ddl_dict = options.as_ddl_dict()

        # Verify the output format
        self.assertTrue(ddl_dict["enable_refresh"])
        self.assertEqual(ddl_dict["refresh_interval_minutes"], 30)
        self.assertEqual(ddl_dict["description"], '"""Test MV"""')

        # Check labels and tags are properly formatted as arrays of tuples
        expected_labels = [("env", "test")]
        expected_tags = [("test-project/team", "data")]

        self.assertEqual(ddl_dict["labels"], expected_labels)
        self.assertEqual(ddl_dict["tags"], expected_tags)

    def test_tags_none_not_included_in_ddl(self):
        """Test that None tags are not included in DDL dict"""
        options = BigQueryOptionsConfig(enable_refresh=True, tags=None)

        ddl_dict = options.as_ddl_dict()

        # tags should not be in the output when None
        self.assertNotIn("tags", ddl_dict)
        self.assertIn("enable_refresh", ddl_dict)

    def test_parse_relation_config_with_resource_tags(self):
        """Test parse_relation_config method handles resource_tags correctly"""
        from unittest.mock import Mock

        # Mock relation config with resource_tags
        mock_relation_config = Mock()
        mock_config = Mock()
        mock_config.extra = {
            "enable_refresh": True,
            "refresh_interval_minutes": 30,
            "description": "Test view",
            "labels": {"env": "test"},
            "resource_tags": {"test-project/team": "data", "test-project/project": "analytics"},
        }
        mock_config.persist_docs = True
        mock_relation_config.config = mock_config

        config_dict = BigQueryOptionsConfig.parse_relation_config(mock_relation_config)

        # Verify resource_tags gets mapped to tags
        self.assertEqual(
            config_dict["tags"], {"test-project/team": "data", "test-project/project": "analytics"}
        )
        self.assertEqual(config_dict["labels"], {"env": "test"})
        self.assertEqual(config_dict["description"], "Test view")
        self.assertTrue(config_dict["enable_refresh"])

    def test_parse_relation_config_without_resource_tags(self):
        """Test parse_relation_config when resource_tags is not present"""
        from unittest.mock import Mock

        # Mock relation config without resource_tags
        mock_relation_config = Mock()
        mock_config = Mock()
        mock_config.extra = {"enable_refresh": False, "labels": {"env": "prod"}}
        mock_config.persist_docs = True
        mock_relation_config.config = mock_config

        config_dict = BigQueryOptionsConfig.parse_relation_config(mock_relation_config)

        # Verify tags is not in the config_dict when resource_tags is not present
        self.assertNotIn("tags", config_dict)
        self.assertEqual(config_dict["labels"], {"env": "prod"})
        self.assertFalse(config_dict["enable_refresh"])


if __name__ == "__main__":
    unittest.main()



from unittest.mock import patch

from dbt.adapters.bigquery.python_submissions import _update_batch_from_config
from google.cloud import dataproc_v1

from .test_bigquery_adapter import BaseTestBigQueryAdapter


# Test application of dataproc_batch configuration to a
# google.cloud.dataproc_v1.Batch object.
# This reuses the machinery from BaseTestBigQueryAdapter to get hold of the
# parsed credentials
class TestConfigureDataprocBatch(BaseTestBigQueryAdapter):
    @patch(
        "dbt.adapters.bigquery.credentials._create_bigquery_defaults",
        return_value=("credentials", "project_id"),
    )
    def test_update_dataproc_serverless_batch(self, mock_get_bigquery_defaults):
        adapter = self.get_adapter("dataproc-serverless-configured")
        mock_get_bigquery_defaults.assert_called_once()

        credentials = adapter.acquire_connection("dummy").credentials
        self.assertIsNotNone(credentials)

        batchConfig = credentials.dataproc_batch
        self.assertIsNotNone(batchConfig)

        raw_batch_config = self.raw_profile["outputs"]["dataproc-serverless-configured"][
            "dataproc_batch"
        ]
        raw_environment_config = raw_batch_config["environment_config"]
        raw_execution_config = raw_environment_config["execution_config"]
        raw_labels: dict[str, any] = raw_batch_config["labels"]
        raw_rt_config = raw_batch_config["runtime_config"]

        raw_batch_config = self.raw_profile["outputs"]["dataproc-serverless-configured"][
            "dataproc_batch"
        ]

        batch = dataproc_v1.Batch()

        batch = _update_batch_from_config(raw_batch_config, batch)

        def to_str_values(d):
            """google's protobuf types expose maps as dict[str, str]"""
            return dict([(k, str(v)) for (k, v) in d.items()])

        self.assertEqual(
            batch.environment_config.execution_config.service_account,
            raw_execution_config["service_account"],
        )
        self.assertFalse(batch.environment_config.execution_config.network_uri)
        self.assertEqual(
            batch.environment_config.execution_config.subnetwork_uri,
            raw_execution_config["subnetwork_uri"],
        )
        self.assertEqual(
            batch.environment_config.execution_config.network_tags,
            raw_execution_config["network_tags"],
        )
        self.assertEqual(batch.labels, to_str_values(raw_labels))
        self.assertEqual(
            batch.runtime_config.properties, to_str_values(raw_rt_config["properties"])
        )

    @patch(
        "dbt.adapters.bigquery.credentials._create_bigquery_defaults",
        return_value=("credentials", "project_id"),
    )
    def test_default_dataproc_serverless_batch(self, mock_get_bigquery_defaults):
        adapter = self.get_adapter("dataproc-serverless-default")
        mock_get_bigquery_defaults.assert_called_once()

        credentials = adapter.acquire_connection("dummy").credentials
        self.assertIsNotNone(credentials)

        batchConfig = credentials.dataproc_batch
        self.assertIsNone(batchConfig)



import pytest

from dbt.contracts.results import RunStatus

from dbt.tests.adapter.incremental.test_incremental_unique_id import SubBaseIncrementalUniqueKey


class IncrementalUniqueKeyFalseyNullsEquals(SubBaseIncrementalUniqueKey):
    def test__bad_unique_key(self, project):
        """expect compilation error from unique key not being a column"""

        (status, exc) = self.fail_to_build_inc_missing_unique_key_column(
            incremental_model_name="not_found_unique_key"
        )

        assert status == RunStatus.Error
        assert "thisisnotacolumn" in exc.lower()

    # test unique_key as list
    def test__empty_unique_key_list(self, project):
        """with no unique keys, seed and model should match"""

        expected_fields = self.get_expected_fields(relation="seed", seed_rows=9)
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="empty_unique_key_list",
            update_sql_file="add_new_rows",
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__one_unique_key(self, project):
        """with one unique key, model will overwrite existing row"""

        expected_fields = self.get_expected_fields(
            relation="one_str__overwrite", seed_rows=8, opt_model_count=1
        )
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="str_unique_key",
            update_sql_file="duplicate_insert",
            opt_model_count=self.update_incremental_model("one_str__overwrite"),
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__bad_unique_key_list(self, project):
        """expect compilation error from unique key not being a column"""

        (status, exc) = self.fail_to_build_inc_missing_unique_key_column(
            incremental_model_name="not_found_unique_key_list"
        )

        assert status == RunStatus.Error
        assert "thisisnotacolumn" in exc.lower()


class IncrementalUniqueKeyTruthyNullsEquals(SubBaseIncrementalUniqueKey):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"flags": {"enable_truthy_nulls_equals_macro": True}}

    # no unique_key test
    def test__no_unique_keys(self, project):
        """with no unique keys, seed and model should match"""

        expected_fields = self.get_expected_fields(relation="seed", seed_rows=9)
        test_case_fields = self.get_test_fields(
            project, seed="seed", incremental_model="no_unique_key", update_sql_file="add_new_rows"
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    # unique_key as str tests
    def test__empty_str_unique_key(self, project):
        """with empty string for unique key, seed and model should match"""

        expected_fields = self.get_expected_fields(relation="seed", seed_rows=9)
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="empty_str_unique_key",
            update_sql_file="add_new_rows",
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__unary_unique_key_list(self, project):
        """with one unique key, model will overwrite existing row"""

        expected_fields = self.get_expected_fields(
            relation="unique_key_list__inplace_overwrite", seed_rows=8, opt_model_count=1
        )
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="unary_unique_key_list",
            update_sql_file="duplicate_insert",
            opt_model_count=self.update_incremental_model("unique_key_list__inplace_overwrite"),
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__duplicated_unary_unique_key_list(self, project):
        """with two of the same unique key, model will overwrite existing row"""

        expected_fields = self.get_expected_fields(
            relation="unique_key_list__inplace_overwrite", seed_rows=8, opt_model_count=1
        )
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="duplicated_unary_unique_key_list",
            update_sql_file="duplicate_insert",
            opt_model_count=self.update_incremental_model("unique_key_list__inplace_overwrite"),
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__trinary_unique_key_list(self, project):
        """with three unique keys, model will overwrite existing row"""

        expected_fields = self.get_expected_fields(
            relation="unique_key_list__inplace_overwrite", seed_rows=8, opt_model_count=1
        )
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="trinary_unique_key_list",
            update_sql_file="duplicate_insert",
            opt_model_count=self.update_incremental_model("unique_key_list__inplace_overwrite"),
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)

    def test__trinary_unique_key_list_no_update(self, project):
        """even with three unique keys, adding distinct rows to seed does not
        cause seed and model to diverge"""

        expected_fields = self.get_expected_fields(relation="seed", seed_rows=9)
        test_case_fields = self.get_test_fields(
            project,
            seed="seed",
            incremental_model="nontyped_trinary_unique_key_list",
            update_sql_file="add_new_rows",
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)


class TestIncrementalUniqueKeyFalseyNullsEquals(IncrementalUniqueKeyFalseyNullsEquals):
    pass


class TestIncrementalUniqueKeyTruthyNullsEquals(IncrementalUniqueKeyTruthyNullsEquals):
    pass



import pytest

from dbt.tests.adapter.incremental.test_incremental_on_schema_change import (
    BaseIncrementalOnSchemaChangeSetup,
    BaseIncrementalOnSchemaChange,
)

from dbt.tests.adapter.incremental.fixtures import (
    _MODELS__A,
    _MODELS__INCREMENTAL_SYNC_ALL_COLUMNS_TARGET,
)

from dbt.tests.util import run_dbt


class TestIncrementalOnSchemaChange(BaseIncrementalOnSchemaChange):
    pass


_MODELS__INCREMENTAL_SYNC_ALL_COLUMNS_DYNAMIC_INSERT_OVERWRITE = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='sync_all_columns',
        partition_by={
            "field": "id",
            "data_type": "int64",
            "range": {
                "start": 1,
                "end": 6,
                "interval": 1
            }
        },
        incremental_strategy='insert_overwrite'
    )
}}

WITH source_data AS (SELECT * FROM {{ ref('model_a') }} )

{% set string_type = 'string' %}

{% if is_incremental() %}

SELECT id,
       cast(field1 as {{string_type}}) as field1,
       cast(field3 as {{string_type}}) as field3, -- to validate new fields
       cast(field4 as {{string_type}}) AS field4 -- to validate new fields

FROM source_data WHERE id > _dbt_max_partition

{% else %}

select id,
       cast(field1 as {{string_type}}) as field1,
       cast(field2 as {{string_type}}) as field2

from source_data where id <= 3

{% endif %}
"""

_MODELS__INCREMENTAL_APPEND_NEW_COLUMNS_DYNAMIC_INSERT_OVERWRITE = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='append_new_columns',
        partition_by={
            "field": "id",
            "data_type": "int64",
            "range": {
                "start": 1,
                "end": 6,
                "interval": 1
            }
        },
        incremental_strategy='insert_overwrite'
    )
}}

{% set string_type = 'string' %}

WITH source_data AS (SELECT * FROM {{ ref('model_a') }} )

{% if is_incremental()  %}

SELECT id,
       cast(field1 as {{string_type}}) as field1,
       cast(field2 as {{string_type}}) as field2,
       cast(field3 as {{string_type}}) as field3,
       cast(field4 as {{string_type}}) as field4
FROM source_data WHERE id > _dbt_max_partition

{% else %}

SELECT id,
       cast(field1 as {{string_type}}) as field1,
       cast(field2 as {{string_type}}) as field2
FROM source_data where id <= 3

{% endif %}
"""

_MODELS__INCREMENTAL_APPEND_NEW_COLUMNS_DYNAMIC_INSERT_OVERWRITE_TARGET = """
{{
    config(materialized='table')
}}

{% set string_type = 'string' %}

with source_data as (

    select * from {{ ref('model_a') }}

)

select id
       ,cast(field1 as {{string_type}}) as field1
       ,cast(field2 as {{string_type}}) as field2
       ,cast(CASE WHEN id <= 3 THEN NULL ELSE field3 END as {{string_type}}) AS field3
       ,cast(CASE WHEN id <= 3 THEN NULL ELSE field4 END as {{string_type}}) AS field4

from source_data
"""

_MODELS__INCREMENTAL_SYNC_ALL_COLUMNS_DYNAMIC_INSERT_OVERWRITE_TARGET = """
{{
    config(materialized='table')
}}

with source_data as (

    select * from {{ ref('model_a') }}

)

{% set string_type = 'string' %}

select id
       ,cast(field1 as {{string_type}}) as field1
       --,field2
       ,cast(case when id <= 3 then null else field3 end as {{string_type}}) as field3
       ,cast(case when id <= 3 then null else field4 end as {{string_type}}) as field4

from source_data
order by id
"""


class TestIncrementalOnSchemaChangeBigQuerySpecific(BaseIncrementalOnSchemaChangeSetup):

    def test_run_incremental_append_new_columns_dynamic_insert_overwrite(self, project):
        select = "model_a incremental_append_new_columns_dynamic_insert_overwrite incremental_append_new_columns_dynamic_insert_overwrite_target"
        compare_source = "incremental_append_new_columns_dynamic_insert_overwrite"
        compare_target = "incremental_append_new_columns_dynamic_insert_overwrite_target"
        self.run_twice_and_assert(select, compare_source, compare_target, project)

    def test_run_incremental_sync_all_columns_dynamic_insert_overwrite(self, project):
        select = "model_a incremental_sync_all_columns_dynamic_insert_overwrite incremental_sync_all_columns_dynamic_insert_overwrite_target"
        compare_source = "incremental_sync_all_columns_dynamic_insert_overwrite"
        compare_target = "incremental_sync_all_columns_dynamic_insert_overwrite_target"
        self.run_twice_and_assert(select, compare_source, compare_target, project)

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental_sync_all_columns_dynamic_insert_overwrite.sql": _MODELS__INCREMENTAL_SYNC_ALL_COLUMNS_DYNAMIC_INSERT_OVERWRITE,
            "incremental_sync_all_columns_dynamic_insert_overwrite_target.sql": _MODELS__INCREMENTAL_SYNC_ALL_COLUMNS_DYNAMIC_INSERT_OVERWRITE_TARGET,
            "incremental_append_new_columns_dynamic_insert_overwrite.sql": _MODELS__INCREMENTAL_APPEND_NEW_COLUMNS_DYNAMIC_INSERT_OVERWRITE,
            "incremental_append_new_columns_dynamic_insert_overwrite_target.sql": _MODELS__INCREMENTAL_APPEND_NEW_COLUMNS_DYNAMIC_INSERT_OVERWRITE_TARGET,
            "model_a.sql": _MODELS__A,
        }


# Test fixtures for special character column names
_MODEL_A_SPECIAL_CHARS = """
{{
    config(materialized='table')
}}

with source_data as (

    select 1 as id, 'bbb' as `select`, 111 as `field-with-dash`
    union all select 2 as id, 'ddd' as `select`, 222 as `field-with-dash`
    union all select 3 as id, 'fff' as `select`, 333 as `field-with-dash`
    union all select 4 as id, 'hhh' as `select`, 444 as `field-with-dash`
    union all select 5 as id, 'jjj' as `select`, 555 as `field-with-dash`
    union all select 6 as id, 'lll' as `select`, 666 as `field-with-dash`

)

select id
       ,`select`
       ,`field-with-dash`

from source_data
"""

_MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='append_new_columns'
    )
}}

WITH source_data AS (SELECT * FROM {{ ref('model_a_special_chars') }} )

{% if is_incremental() %}

SELECT id,
       `select`,
       `field-with-dash`
FROM source_data WHERE id NOT IN (SELECT id from {{ this }} )

{% else %}

SELECT id,
       `select`
FROM source_data where id <= 3

{% endif %}
"""

_MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS_TARGET = """
{{
    config(materialized='table')
}}

with source_data as (

    select * from {{ ref('model_a_special_chars') }}

)

select id
       ,`select`
       ,CASE WHEN id <= 3 THEN NULL ELSE `field-with-dash` END AS `field-with-dash`

from source_data
"""


class TestIncrementalOnSchemaChangeSpecialChars(BaseIncrementalOnSchemaChangeSetup):
    """Test incremental models with special character column names for BigQuery

    Tests column quoting with SQL keywords and dashes in column names.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_a_special_chars.sql": _MODEL_A_SPECIAL_CHARS,
            "incremental_append_new_special_chars.sql": _MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS,
            "incremental_append_new_special_chars_target.sql": _MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS_TARGET,
        }

    def test_incremental_append_new_columns_with_special_characters(self, project):
        """Test that incremental models work correctly with special character column names when adding new columns"""

        # First run - creates initial table
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_chars incremental_append_new_special_chars incremental_append_new_special_chars_target",
            ]
        )

        # Second run - should append new columns with special characters
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_chars incremental_append_new_special_chars incremental_append_new_special_chars_target",
            ]
        )

        # Verify results match expected output
        from dbt.tests.util import check_relations_equal

        check_relations_equal(
            project.adapter,
            [
                "incremental_append_new_special_chars",
                "incremental_append_new_special_chars_target",
            ],
        )



from dbt.adapters.bigquery.dataset import add_access_entry_to_dataset, is_access_entry_in_dataset
from dbt.adapters.bigquery import BigQueryRelation

from google.cloud.bigquery import Dataset, AccessEntry, DatasetReference


def test_add_access_entry_to_dataset_updates_dataset():
    database = "someDb"
    dataset = "someDataset"
    entity = BigQueryRelation.from_dict(
        {
            "type": None,
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "my_table",
            },
            "quote_policy": {"identifier": False},
        }
    ).to_dict()
    dataset_ref = DatasetReference(project=database, dataset_id=dataset)
    dataset = Dataset(dataset_ref)
    access_entry = AccessEntry(None, "table", entity)
    dataset = add_access_entry_to_dataset(dataset, access_entry)
    assert is_access_entry_in_dataset(dataset, access_entry)


def test_add_access_entry_to_dataset_updates_with_pre_existing_entries():
    database = "someOtherDb"
    dataset = "someOtherDataset"
    entity_2 = BigQueryRelation.from_dict(
        {
            "type": None,
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "some_other_view",
            },
            "quote_policy": {"identifier": False},
        }
    ).to_dict()
    dataset_ref = DatasetReference(project=database, dataset_id=dataset)
    dataset = Dataset(dataset_ref)
    initial_entry = AccessEntry(None, "view", entity_2)
    initial_entry._properties.pop("role")
    dataset.access_entries = [initial_entry]
    access_entry = AccessEntry(None, "view", entity_2)
    dataset = add_access_entry_to_dataset(dataset, access_entry)
    assert len(dataset.access_entries) == 2


def test_is_access_entry_in_dataset_returns_true_if_entry_in_dataset():
    database = "someDb"
    dataset = "someDataset"
    entity = BigQueryRelation.from_dict(
        {
            "type": None,
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "my_table",
            },
            "quote_policy": {"identifier": False},
        }
    ).to_dict()
    dataset_ref = DatasetReference(project=database, dataset_id=dataset)
    dataset = Dataset(dataset_ref)
    access_entry = AccessEntry(None, "table", entity)
    dataset = add_access_entry_to_dataset(dataset, access_entry)
    assert is_access_entry_in_dataset(dataset, access_entry)


def test_is_access_entry_in_dataset_returns_false_if_entry_not_in_dataset():
    database = "someDb"
    dataset = "someDataset"
    entity = BigQueryRelation.from_dict(
        {
            "type": None,
            "path": {
                "database": "test-project",
                "schema": "test_schema",
                "identifier": "my_table",
            },
            "quote_policy": {"identifier": False},
        }
    ).to_dict()
    dataset_ref = DatasetReference(project=database, dataset_id=dataset)
    dataset = Dataset(dataset_ref)
    access_entry = AccessEntry(None, "table", entity)
    assert not is_access_entry_in_dataset(dataset, access_entry)



from dataclasses import dataclass, replace
from itertools import product
from typing import Optional

from dbt.tests.util import run_dbt
import pytest

from dbt.adapters.bigquery import constants

from tests.functional.adapter.relation_tests import models
from tests.functional.utils import query_relation_type, update_model


@dataclass
class Model:
    model: str
    relation_type: str
    table_format: Optional[str] = constants.INFO_SCHEMA_TABLE_FORMAT
    is_incremental: Optional[bool] = False

    @property
    def name(self) -> str:
        if self.is_incremental:
            name = f"{self.relation_type}_{self.table_format}_incremental"
        else:
            name = f"{self.relation_type}_{self.table_format}"
        return name

    @property
    def is_standard_table(self) -> bool:
        return self.relation_type == "table" and not self.is_incremental


@dataclass
class Scenario:
    initial: Model
    final: Model

    @property
    def name(self) -> str:
        return f"REPLACE_{self.initial.name}__WITH_{self.final.name}"

    @property
    def error_message(self) -> str:
        return f"Failed when migrating from: {self.initial.name} to: {self.final.name}"


relations = [
    Model(models.VIEW, "view"),
    Model(models.TABLE, "table"),
    Model(models.MATERIALIZED_VIEW, "materialized_view"),
    Model(
        models.INCREMENTAL_TABLE, "table", constants.INFO_SCHEMA_TABLE_FORMAT, is_incremental=True
    ),
    Model(models.ICEBERG_TABLE, "table", constants.ICEBERG_TABLE_FORMAT),
    Model(
        models.INCREMENTAL_ICEBERG_TABLE,
        "table",
        constants.ICEBERG_TABLE_FORMAT,
        is_incremental=True,
    ),
]


scenarios = [Scenario(*scenario) for scenario in product(relations, relations)]


# replace the storage_uri with the specific scenario name to avoid naming conflicts on the same initial model
for scenario in scenarios:
    if scenario.initial.table_format == constants.ICEBERG_TABLE_FORMAT:
        new_model = scenario.initial.model.replace("||storage_uri||", scenario.name)
        scenario.initial = replace(scenario.initial, model=new_model)
    if scenario.final.table_format == constants.ICEBERG_TABLE_FORMAT:
        new_model = scenario.final.model.replace("||storage_uri||", scenario.name)
        scenario.final = replace(scenario.final, model=new_model)


def requires_full_refresh(scenario) -> bool:
    """
    Some scenarios require a full refresh due to a combination of dbt-bigquery's limitations:
    - we have not handled table <> view well
    and BigQuery's limitations:
    https://cloud.google.com/bigquery/docs/iceberg-tables#limitations
    - Iceberg tables cannot be renamed
    """
    return any(
        [
            # dbt does not support replacing non-views with a view
            scenario.initial.relation_type != "view"
            and scenario.final.relation_type == "view",
        ]
    )


def unsupported(scenario) -> bool:
    """
    Some scenarios are not supported at all, even when specifying `--full-refresh`
    """
    return any(
        [
            # dbt does not support replacing a materialized view with an incremental table
            scenario.initial.relation_type == "materialized_view"
            and scenario.final.is_incremental,
            # dbt does not support replacing Iceberg tables with materialized views
            # this is likely due to Iceberg tables not being able to be renamed
            scenario.initial.table_format == constants.ICEBERG_TABLE_FORMAT
            and scenario.final.relation_type == "materialized_view",
        ]
    )


class TestRelationTypeChange:

    @staticmethod
    def include(scenario) -> bool:
        return all(
            [
                not requires_full_refresh(scenario),
                not unsupported(scenario),
            ]
        )

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            f"{scenario.name}.sql": scenario.initial.model
            for scenario in scenarios
            if self.include(scenario)
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        for scenario in scenarios:
            if self.include(scenario):
                update_model(project, scenario.name, scenario.final.model)
        # allow for dbt to fail so that we can see which scenarios pass and which scenarios fail
        try:
            run_dbt(["run"], expect_pass=False)
        except Exception:
            pass

    @pytest.mark.parametrize("scenario", scenarios, ids=[scenario.name for scenario in scenarios])
    def test_replace(self, project, scenario):
        if self.include(scenario):
            relation_type = query_relation_type(project, scenario.name)
            assert (
                relation_type.casefold() == scenario.final.relation_type.casefold()
            ), scenario.error_message
        else:
            pytest.skip()


class TestRelationTypeChangeFullRefreshRequired(TestRelationTypeChange):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"full_refresh": True}}

    @staticmethod
    def include(scenario) -> bool:
        return all(
            [
                requires_full_refresh(scenario),
                not unsupported(scenario),
            ]
        )



import pytest
from dbt.tests.fixtures.project import write_project_files


models__view_2_sql = """
{%- if target.type == 'bigquery' -%}
  {{ config(project=var('alternate_db')) }}
{%- else -%}
  {{ config(database=var('alternate_db')) }}
{%- endif -%}
select * from {{ ref('seed') }}

"""

models__view_1_sql = """
{#
    We are running against a database that must be quoted.
    These calls ensure that we trigger an error if we're failing to quote at parse-time
#}
{% do adapter.already_exists(this.schema, this.table) %}
{% do adapter.get_relation(this.database, this.schema, this.table) %}
select * from {{ ref('seed') }}

"""

models__subfolder__view_4_sql = """
{{
    config(database=var('alternate_db'))
}}

select * from {{ ref('seed') }}

"""

models__subfolder__view_3_sql = """
select * from {{ ref('seed') }}

"""

seeds__seed_csv = """id,name
1,a
2,b
3,c
4,d
5,e
"""


@pytest.fixture(scope="class")
def models():
    return {
        "view_2.sql": models__view_2_sql,
        "view_1.sql": models__view_1_sql,
        "subfolder": {
            "view_4.sql": models__subfolder__view_4_sql,
            "view_3.sql": models__subfolder__view_3_sql,
        },
    }


@pytest.fixture(scope="class")
def seeds():
    return {"seed.csv": seeds__seed_csv}


@pytest.fixture(scope="class")
def project_files(
    project_root,
    models,
    seeds,
):
    write_project_files(project_root, "models", models)
    write_project_files(project_root, "seeds", seeds)



import platform

import time

import os
import signal
import subprocess

import pytest

from dbt.tests.util import get_connection

_SEED_CSV = """
id, name, astrological_sign, moral_alignment
1, Alice, Aries, Lawful Good
2, Bob, Taurus, Neutral Good
3, Thaddeus, Gemini, Chaotic Neutral
4, Zebulon, Cancer, Lawful Evil
5, Yorick, Leo, True Neutral
6, Xavier, Virgo, Chaotic Evil
7, Wanda, Libra, Lawful Neutral
"""

_LONG_RUNNING_MODEL_SQL = """
    {{ config(materialized='table') }}
    with array_1 as (
    select generated_ids from UNNEST(GENERATE_ARRAY(1, 200000)) AS generated_ids
    ),
    array_2 as (
    select generated_ids from UNNEST(GENERATE_ARRAY(2, 200000)) AS generated_ids
    )

    SELECT array_1.generated_ids
    FROM array_1
    LEFT JOIN array_1 as jnd on 1=1
    LEFT JOIN array_2 as jnd2 on 1=1
    LEFT JOIN array_1 as jnd3 on jnd3.generated_ids >= jnd2.generated_ids
"""


def _get_info_schema_jobs_query(project_id, dataset_id, table_id):
    """
    Running this query requires roles/bigquery.resourceViewer on the project,
    see: https://cloud.google.com/bigquery/docs/information-schema-jobs#required_role
    :param project_id:
    :param dataset_id:
    :param table_id:
    :return: a single job id that matches the model we tried to create and was cancelled
    """
    return f"""
        SELECT job_id
        FROM `region-us`.`INFORMATION_SCHEMA.JOBS_BY_PROJECT`
        WHERE creation_time > TIMESTAMP_SUB(CURRENT_TIMESTAMP(), INTERVAL 5 HOUR)
        AND statement_type = 'CREATE_TABLE_AS_SELECT'
        AND state = 'DONE'
        AND job_id IS NOT NULL
        AND project_id = '{project_id}'
        AND error_result.reason = 'stopped'
        AND error_result.message = 'Job execution was cancelled: User requested cancellation'
        AND destination_table.table_id = '{table_id}'
        AND destination_table.dataset_id = '{dataset_id}'
    """


def _run_dbt_in_subprocess(project, dbt_command):

    run_dbt_process = subprocess.Popen(
        [
            "dbt",
            dbt_command,
            "--profiles-dir",
            project.profiles_dir,
            "--project-dir",
            project.project_root,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
        env=os.environ.copy(),
    )
    std_out_log = ""
    while True:
        std_out_line = run_dbt_process.stdout.readline().decode("utf-8")
        std_out_log += std_out_line
        if std_out_line != "":
            print(std_out_line)
            if "1 of 1 START" in std_out_line:
                time.sleep(1)
                run_dbt_process.send_signal(signal.SIGINT)

        if run_dbt_process.poll():
            break

    return std_out_log


def _get_job_id(project, table_name):
    # Because we run this in a subprocess we have to actually call Bigquery and look up the job id
    with get_connection(project.adapter):
        job_id = project.run_sql(
            _get_info_schema_jobs_query(project.database, project.test_schema, table_name)
        )

    return job_id


@pytest.mark.skipif(
    platform.system() == "Windows", reason="running signt is unsupported on Windows."
)
class TestBigqueryCancelsQueriesOnKeyboardInterrupt:
    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        return {
            "model.sql": _LONG_RUNNING_MODEL_SQL,
        }

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {
            "seed.csv": _SEED_CSV,
        }

    def test_bigquery_cancels_queries_for_model_on_keyboard_interrupt(self, project):
        std_out_log = _run_dbt_in_subprocess(project, "run")

        assert "CANCEL query model.test.model" in std_out_log
        assert len(_get_job_id(project, "model")) == 1

    @pytest.mark.skip(reason="cannot reliably cancel seed queries in time")
    def test_bigquery_cancels_queries_for_seed_on_keyboard_interrupt(self, project):
        std_out_log = _run_dbt_in_subprocess(project, "seed")

        assert "CANCEL query seed.test.seed" in std_out_log
        # we can't assert the job id since we can't kill the seed process fast enough to cancel it



import requests
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Optional
from datetime import datetime, timedelta, timezone
from google.auth.identity_pool import SubjectTokenSupplier
from dbt.adapters.exceptions import FailedToConnectError
from google.auth.external_account import SupplierContext
from google.auth.transport import Request
from dbt_common.exceptions import DbtRuntimeError


class TokenServiceBase(ABC):
    def __init__(self, token_endpoint: Dict[str, Any]):
        expected_keys = {"type", "request_url", "request_data"}
        for key in expected_keys:
            if key not in token_endpoint:
                raise FailedToConnectError(f"Missing required key in token_endpoint: '{key}'")

        self.type: str = token_endpoint["type"]
        self.url: str = token_endpoint["request_url"]
        self.data: str = token_endpoint["request_data"]

        self.other_params = {k: v for k, v in token_endpoint.items() if k not in expected_keys}

    @abstractmethod
    def build_header_payload(self) -> Dict[str, Any]:
        pass

    def handle_request(self) -> requests.Response:
        """
        Handles the request with rate limiting and error handling.
        """
        response = requests.post(self.url, headers=self.build_header_payload(), data=self.data)

        if response.status_code == 429:
            raise DbtRuntimeError(
                "Rate limit on identity provider's token dispatch has been reached. "
                "Consider increasing your identity provider's refresh token rate or "
                "lower dbt's maximum concurrent thread count."
            )

        response.raise_for_status()
        return response


class EntraIdpTokenService(TokenServiceBase):
    """
    formatted based on docs: https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow#get-a-token
    """

    def build_header_payload(self) -> Dict[str, Any]:
        return {
            "accept": "application/json",
            "content-type": "application/x-www-form-urlencoded",
        }


class TokenServiceType(Enum):
    ENTRA = "entra"


class EntraTokenSupplier(SubjectTokenSupplier):
    """
    A SubjectTokenSupplier that fetches tokens from Entra ID.

    This is an implementation of the SubjectTokenSupplier interface that can be specified as the
    credential_source field in a Google Cloud Application Default Credentials (ADC) config file.

    https://googleapis.dev/python/google-auth/latest/reference/google.auth.identity_pool.html#google.auth.identity_pool.SubjectTokenSupplier
    """

    def __init__(self, token_endpoint: Dict[str, Any]):
        self.token_service = EntraIdpTokenService(token_endpoint)
        self._cached_token: Optional[str] = None
        self._token_expiry: Optional[datetime] = None
        # Add a 5-minute buffer before actual expiry to ensure we don't use an expired token
        self._expiry_buffer = timedelta(minutes=5)

    def _is_token_valid(self) -> bool:
        """Check if the cached token is still valid."""
        if not self._token_expiry:
            return False
        return datetime.now(timezone.utc).replace(tzinfo=None) < (
            self._token_expiry - self._expiry_buffer
        )

    def _fetch_new_token(self) -> str:
        """Fetch a new token from Entra ID."""
        response = self.token_service.handle_request()
        token_data = response.json()

        if "access_token" not in token_data:
            raise FailedToConnectError(
                "access_token missing from Idp token request. Please confirm correct "
                "configuration of the token_endpoint field in profiles.yml and that your "
                "Idp can obtain an OIDC-compliant access token."
            )

        # Extract expiration time from token response
        # Default to 1 hour if not specified
        expires_in = token_data.get("expires_in", 3600)
        self._token_expiry = datetime.now(timezone.utc).replace(tzinfo=None) + timedelta(
            seconds=expires_in
        )
        self._cached_token = token_data["access_token"]

        return token_data["access_token"]

    def get_subject_token(self, context: SupplierContext, request: Request) -> str:
        """
        Fetch and return a subject token from Entra ID.

        Args:
            context: The context object containing information about the requested audience and subject token type
            request: The object used to make HTTP requests

        Returns:
            str: The access token from Entra ID

        Raises:
            FailedToConnectError: If token acquisition fails or response is invalid
        """
        if self._cached_token and self._is_token_valid():
            return self._cached_token

        return self._fetch_new_token()


def create_token_supplier(token_endpoint: Dict[str, Any]) -> SubjectTokenSupplier:
    if (service_type := token_endpoint.get("type")) is None:
        raise FailedToConnectError("Missing required key in token_endpoint: 'type'")

    if service_type == TokenServiceType.ENTRA.value:
        return EntraTokenSupplier(token_endpoint)
    else:
        raise ValueError(
            f"Unsupported identity provider type: {service_type}. Only 'entra' is supported at this time"
        )



from dataclasses import dataclass
from typing import Optional

from dbt.adapters.catalogs import CatalogRelation

from dbt.adapters.bigquery import constants


@dataclass
class BigQueryCatalogRelation(CatalogRelation):
    catalog_type: str = constants.DEFAULT_INFO_SCHEMA_CATALOG.catalog_type
    catalog_name: Optional[str] = constants.DEFAULT_INFO_SCHEMA_CATALOG.name
    table_format: Optional[str] = constants.INFO_SCHEMA_TABLE_FORMAT
    file_format: Optional[str] = constants.INFO_SCHEMA_FILE_FORMAT
    external_volume: Optional[str] = None
    storage_uri: Optional[str] = None
